import numpy as np

from molesq import Transformer
from .reference_impl import moving_least_squares_affine_vectorized


def test_importable():
    import molesq

    assert molesq.__version__


def test_instantiates(brain_landmarks):
    assert Transformer(*brain_landmarks)


def test_against_ref(brain_landmarks, neurons):
    n = neurons[0]
    t = Transformer(*brain_landmarks)
    test = t.transform(n)
    ref = moving_least_squares_affine_vectorized(n, *brain_landmarks)
    assert np.allclose(test, ref)


def test_ref_works(brain_landmarks, neurons):
    n = neurons[0]
    # t = Transformer(*brain_landmarks)
    # test = t.transform(n)
    ref = moving_least_squares_affine_vectorized(n, *brain_landmarks)
    assert n.shape == ref.shape
    assert not np.allclose(n, ref)



from pathlib import Path

import numpy as np

import pytest

TEST_DIR = Path(__file__).absolute().parent
PROJECT_DIR = TEST_DIR.parent
DATA_DIR = PROJECT_DIR / "data"


def read_swc(fpath):
    return np.genfromtxt(
        fpath,
        usecols=(2, 3, 4),
        delimiter=" ",
    )


@pytest.fixture(scope="session")
def _neurons():
    return [read_swc(f) for f in sorted(DATA_DIR.glob("*.swc"))]


@pytest.fixture
def neurons(_neurons):
    return _neurons[0].copy(), _neurons[1].copy()


@pytest.fixture(scope="session")
def _brain_landmarks():
    vals = np.genfromtxt(
        DATA_DIR / "Brain_Lineage_Landmarks_EMtoEM_ProjectSpace.csv",
        skip_header=1,
        usecols=(1, 2, 3, 4, 5, 6),
        delimiter=",",
    )
    left = vals[:, :3]
    right = vals[:, 3:]
    from_cp = np.vstack([left, right])
    to_cp = np.vstack([right, left])
    return from_cp, to_cp


@pytest.fixture
def brain_landmarks(_brain_landmarks):
    return _brain_landmarks[0].copy(), _brain_landmarks[1].copy()



from .version import version as __version__  # noqa: F401
from .version import version_tuple as __version_info__  # noqa: F401

from .transform import Strategy, Transformer

__all__ = ["Strategy", "Transformer"]



import sys

import scipy as sp
import numpy as np


def moving_least_squares_affine_vectorized(vs, ps, qs):
    """
    Compute the 3d affine deformation based on Schaefer et al 2006.

    This implementation from
    https://github.com/ceesem/catalysis/blob/master/catalysis/transform.py

    Parameters
    ----------
    vs : Npoint x 3 numpy array
        Block of points to be transformed

    ps : Nlandmark x 3 numpy array
        Block of landmarks in the starting coordinates

    qs : Nlandmark x 3 numpy array
        Block of landmarks in the transformed coordinates, matched indices.

    Returns
    -------

    vprime : Npoint x 3 numpy array
        Block of points in the transformed coordinates.
    """
    Nlandmark = len(ps)
    Npoint = len(vs)

    # Reshape into arrays with consistent indices
    ps = ps.ravel().reshape(1, 3, Nlandmark, 1, order="F")
    qs = qs.ravel().reshape(1, 3, Nlandmark, 1, order="F")
    vs = vs.ravel().reshape(1, 3, 1, Npoint, order="F")

    ds = (
        sp.spatial.distance.cdist(
            ps.ravel("F").reshape(Nlandmark, 3),
            vs.ravel("F").reshape(Npoint, 3),
            "sqeuclidean",
        ).reshape(1, 1, Nlandmark, Npoint)
        + sys.float_info.epsilon
    )
    ws = 1 / ds

    wi_norm_inv = 1 / np.sum(ws, axis=2)
    pstar = np.einsum("ijl,ijkl,ijkl->ijl", wi_norm_inv, ws, ps).reshape(
        1, 3, 1, Npoint
    )
    qstar = np.einsum("ijl,ijkl,ijkl->ijl", wi_norm_inv, ws, qs).reshape(
        1, 3, 1, Npoint
    )

    phat = ps - pstar
    qhat = qs - qstar

    vminusp = vs - pstar

    Y = np.einsum("ijkl,mikl,mjkl->ijl", ws, phat, phat).reshape(3, 3, 1, Npoint)
    Yinv = np.zeros((3, 3, 1, Npoint))
    for i in range(Npoint):
        Yinv[:, :, 0, i] = np.linalg.inv(Y[:, :, 0, i])

    Z = np.einsum("ijkl,mikl,mjkl->ijl", ws, phat, qhat).reshape(3, 3, 1, Npoint)

    vprime = np.einsum("iakl,abkl,bjkl->ijkl", vminusp, Yinv, Z) + qstar
    vprime = vprime.ravel("F").reshape(Npoint, 3)
    return vprime



import sys

import numpy as np
from scipy.spatial.distance import cdist

try:
    from enum import StrEnum
except ImportError:
    from backports.strenum import StrEnum


class Strategy(StrEnum):
    AFFINE = "affine"
    # SIMILARITY = "similarity"
    # RIGID = "rigid"


ORDER = "F"


def reshape_landmarks(locations: np.ndarray) -> np.ndarray:
    """Reshape NxD landmark array into 1xDxNx1.

    Where D is the dimensionality, and N the number of landmarks.

    Parameters
    ----------
    locations

    Returns
    -------
    reshaped array
    """
    n_locations, n_dimensions = locations.shape
    return locations.ravel().reshape(1, n_dimensions, n_locations, 1, order=ORDER)


class Transformer:
    def __init__(self, from_landmarks, to_landmarks) -> None:
        from_arr = np.asarray(from_landmarks)
        to_arr = np.asarray(to_landmarks)

        if from_arr.shape != to_arr.shape:
            raise ValueError("Landmarks must have the same shape")
        if len(from_arr.shape) != 2:
            raise ValueError("Landmarks must be 2D array")

        self.n_landmarks, self.dimensions = from_arr.shape

        self.from_landmarks = reshape_landmarks(from_arr)
        self.to_landmarks = reshape_landmarks(to_arr)

    def _transform_affine(self, locs: np.ndarray, orig_cp, deformed_cp) -> np.ndarray:
        n_locs = locs.shape[-1]

        # Pairwise distances between original control points and locations to transform
        # reshaped to 1,1,N_cp,N_l
        # jittered to avoid 0s
        distances = (
            cdist(
                orig_cp.ravel(ORDER).reshape(self.n_landmarks, self.dimensions),
                locs.ravel(ORDER).reshape(n_locs, self.dimensions),
                "sqeuclidean",
            ).reshape(1, 1, self.n_landmarks, n_locs)
            + sys.float_info.epsilon
        )
        weights = 1 / distances

        weights_inverse_norm = 1 / np.sum(weights, axis=2)

        # weighted centroids
        orig_star = np.einsum(
            "ijl,ijkl,ijkl->ijl", weights_inverse_norm, weights, orig_cp
        ).reshape(1, self.dimensions, 1, n_locs)
        deformed_star = np.einsum(
            "ijl,ijkl,ijkl->ijl", weights_inverse_norm, weights, deformed_cp
        ).reshape(1, self.dimensions, 1, n_locs)

        # distance to weighted centroids
        orig_hat = orig_cp - orig_star
        deformed_hat = deformed_cp - deformed_star

        Y = np.einsum("ijkl,mikl,mjkl->ijl", weights, orig_hat, orig_hat).reshape(
            self.dimensions, self.dimensions, 1, n_locs
        )

        Y_inv = np.zeros((self.dimensions, self.dimensions, 1, n_locs))
        for i in range(n_locs):
            Y_inv[:, :, 0, i] = np.linalg.inv(Y[:, :, 0, i])

        Z = np.einsum("ijkl,mikl,mjkl->ijl", weights, orig_hat, deformed_hat).reshape(
            self.dimensions, self.dimensions, 1, n_locs
        )

        vprime = (
            np.einsum("iakl,abkl,bjkl->ijkl", locs - orig_star, Y_inv, Z)
            + deformed_star
        )
        vprime = vprime.ravel(ORDER).reshape(n_locs, self.dimensions)

        return vprime

    def transform(
        self,
        locations,
        inverse=False,
        strategy=Strategy.AFFINE,
    ) -> np.ndarray:
        locs = np.asarray(locations)

        orig_cp = self.from_landmarks
        deformed_cp = self.to_landmarks

        if inverse:
            orig_cp, deformed_cp = deformed_cp, orig_cp

        # Reshape into arrays with consistent indices
        locs = locs.ravel().reshape(1, self.dimensions, 1, len(locs), order=ORDER)

        if strategy == Strategy.AFFINE:
            return self._transform_affine(locs, orig_cp, deformed_cp)
        else:
            raise ValueError(f"Unimplemented/ unknown strategy {strategy}")



from pathlib import Path
from setuptools import setup, find_packages

with open(Path(__file__).resolve().parent / "README.md") as f:
    readme = f.read()

setup(
    name="molesq",
    url="https://github.com/clbarnes/molesq",
    author="Chris L. Barnes",
    description="Implementation of moving least squares",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(include=["molesq"]),
    install_requires=["numpy", "scipy", "backports.strenum; python_version < '3.10'"],
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
)
